﻿//'*****************************************************************
// *                                                               *
// *       Copyright (c) Fast Corporation, 2007                    *
// *                                                               *
// *   All Rights Reserved.  Unpublished rights  reserved  under   *
// *   the copyright laws of the Japan.                            *
// *                                                               *
// *   The software contained on this media  is  proprietary  to   *
// *   and  embodies  the  confidential  technology  of  Fast      *
// *   Corporation.  Possession, use,  duplication  or             *
// *   dissemination of the software and media is authorized only  *
// *   pursuant to a valid written license from Fast Corporation.  *
// *                                                               *
// *****************************************************************/

// -------------------------------------------------------------------
// FVX Sample Program
// メインダイアログ
//
// 作成者 : R.Nagashima
//
// 履歴 : 2007/03/13 初版
// -------------------------------------------------------------------
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Threading;
using System.IO;
//FVXデータ構造の使用宣言
using FVDATASLib;


namespace CircleEdgeDistance
{
	public partial class frmMain : Form
	{
		public frmMain()
		{
			InitializeComponent();
		}
		
		//π
		const double PI = 3.14159265358979;
		//処理範囲
		FvRectangle m_ExeRegion = new FvRectangle();
		// ビデオデバイスがオープンされているか
		private bool m_blnVideoOpened = false;
		// ビデオ設定ファイル名（フルパス）
		private string m_strVideoConfFilePath = "";
		// 初期ディレクトリ
		private string m_strDirectory;
		//エッジ検出線
		private FvCircle m_EdgeScanCircle = new FvCircle();
		
		private void frmMain_Load(object sender, EventArgs e)
		{
			frmSelectvideoConfFile fVideoConf = new frmSelectvideoConfFile();
			frmLicense fLicense = new frmLicense();
			
			// 変数初期化
			m_blnVideoOpened = false;
			btnGrab.Enabled = false;
			
			cmbEdgeColor.Items.Add("白→黒"); //WHITE_TO_BLACK
			cmbEdgeColor.Items.Add("黒→白"); //BLACK_TO_WHITE
			cmbEdgeColor.Items.Add("無極性"); //ALL_COLOR
			cmbEdgeColor.SelectedIndex = 2;

			cmbEdgePoint.Items.Add("最大点");  //POINT_PEAK 微分値が最大の位置をエッジとします｡
			cmbEdgePoint.Items.Add("最初の点"); //POINT_1ST しきい値を一番最初に越えた位置をエッジとします｡
			cmbEdgePoint.SelectedIndex = 0;

			cmbEdgeStartPos.Items.Add("中心→外周"); //FROM_START 始点位置からエッジ走査します｡
			cmbEdgeStartPos.Items.Add("外周→中心"); //FROM_END 終点位置からエッジ走査します｡
			cmbEdgeStartPos.SelectedIndex = 0;

			cmbOrigin.Items.Add("検出円の中心");
			cmbOrigin.Items.Add("エッジから求めた円の中心");
			cmbOrigin.SelectedIndex = 1;

			//エッジ検出円の初期値設定
			m_EdgeScanCircle.x = 290;
			m_EdgeScanCircle.y = 240;
			m_EdgeScanCircle.radius = 120;

			//画面の生成
			axFvxDisplayEx1.CreateDisplay();

			//処理範囲の初期値設定(画像の全面に設定）
			m_ExeRegion.xs = -1;
			m_ExeRegion.xe = -1;
			m_ExeRegion.ys = -1;
			m_ExeRegion.ye = -1;

			//作業ディレクトリ退避
			m_strDirectory = Directory.GetCurrentDirectory();

			//使用許諾の確認
			if (fLicense.ShowDialog() == DialogResult.OK)
			{
				//ビデオ設定ファイルの選択
				if (fVideoConf.ShowDialog() == DialogResult.OK)
					m_strVideoConfFilePath = fVideoConf.strVideoConfFilePath;

				if (fVideoConf.chkNotUseCamera.Checked == false)
				{
					int IntErr;

					//ビデオ設定ファイルをコンポーネントに設定
					axFvxVideo1.VideoConfFile = m_strVideoConfFilePath;

					//デバイスオープン
					try
					{
						axFvxVideo1.OpenDevice();
					}
					catch (Exception)
					{
						axFvxVideo1.GetErrReport(out IntErr);
						MessageBox.Show("ビデオデバイスのオープンに失敗しました｡ ErrCode= " + IntErr.ToString() + "\r\n画像ファイルのみの利用になります。");
						return;
					}

					//VideoConfFileの変更後、OpenDeviceを行うと、トリガーモードの変更が伴います
					//このため、弊社キャプチャボードFVC02では350msのセットアップタイムが必要になります｡
					//本プログラムでは簡略化のために、一律に350msのセットアップタイムを確保していますが、
					//お使いのキャプチャボードに合わせて変更してください。
					Thread.Sleep(350);

					//GrabSync/GrabASyncメソッドではImegeサイズの変更を行いませんので、
					//あらかじめChangeImageSizeメソッドにてImageのサイズを変更してください
					//（取り込みサイズとImageのサイズが違う場合エラーになります）
					axFvxVideo1.ChangeImageSize(axFvxImage1.GetOcx());
					try
					{
						// 同期取込
						axFvxVideo1.GrabSync(axFvxImage1.GetOcx());
					}
					catch (Exception)
					{
						axFvxVideo1.GetErrReport(out IntErr);
						MessageBox.Show("ビデオの取込に失敗しました｡ ErrCode= " + IntErr.ToString() + "\r\n画像ファイルのみの利用になります。");
						return;
					}

					//ビデオデバイスの状態を保存
					m_blnVideoOpened = true;
					btnGrab.Enabled = true;

					//画像を表示
					axFvxDisplayEx1.DrawImage(axFvxImage1.GetOcx());
				}
				ApplyChanges();
				DrawEdgeScanLine();
			}
			else
			{
				//終了
				Application.Exit();
				return;
			}			
		}

		//終了時の処理
		private void frmMain_FormClosed(object sender, FormClosedEventArgs e)
		{
			if (m_blnVideoOpened)
			{
				// ビデオデバイスがオープンされていた場合はクローズします｡
				axFvxVideo1.CloseDevice();
			}
		}

		//カメラ画像取込
		private void btnGrab_Click(object sender, EventArgs e)
		{
			//画像ファイル読み込みによって、画像サイズが初期化時から
			//変更されている可能性があるため、ChangeImageSizeを行います。
			//画像サイズが変更される可能性が無い場合は、初期化時に一度だけ行えば問題有りません。
			axFvxVideo1.ChangeImageSize(axFvxImage1.GetOcx());

			try
			{
				//同期取込
				axFvxVideo1.GrabSync(axFvxImage1.GetOcx());
			}
			catch (Exception)
			{
				int IntErr;
				axFvxVideo1.GetErrReport(out IntErr);
				MessageBox.Show("ビデオの取込に失敗しました｡ ErrCode= " + IntErr.ToString());
			}

			//取り込んだ画像を画面に表示
			axFvxDisplayEx1.DrawImage(axFvxImage1.GetOcx());
		}

		//画像ファイル読込
		private void btnLoadImage_Click(object sender, EventArgs e)
		{
			// ダイアログにて画像ファイルを選択します｡
			openFileDialog1.InitialDirectory = m_strDirectory;
			openFileDialog1.Filter = "bitmap file(*.bmp)|*.bmp";
			openFileDialog1.DefaultExt = "*.bmp";
			openFileDialog1.ShowDialog();

			if (openFileDialog1.FileName.Length > 0)
			{
				//FileAccessコンポーネントを用いて、
				//ファイルからImageに画像を読み込みます｡
				axFvxFileAccess1.FilePath = openFileDialog1.FileName;

				try
				{
					axFvxFileAccess1.LoadBitmap(axFvxImage1.GetOcx());
				}
				catch (Exception)
				{
					int IntErr = 0;
					axFvxFileAccess1.GetErrReport(out IntErr);
					MessageBox.Show("読込みに失敗しました｡ ErrCode= " + IntErr.ToString());
					return;
				}

				//画像タイプチェック
				if ((axFvxImage1.BitPerPixel != 8) || (axFvxImage1.ImageType != FVXCOMIMAGELib.FVXIMAGE_IMAGE_TYPE.imageGray))
				{
					MessageBox.Show("8bitグレイの画像ではありません｡");
					axFvxImage1.BitPerPixel = 8;
					axFvxImage1.ImageType = FVXCOMIMAGELib.FVXIMAGE_IMAGE_TYPE.imageGray;
					axFvxImage1.ImageWidth = 512;
					axFvxImage1.ImageHeight = 480;
				}
				axFvxDisplayEx1.DrawImage(axFvxImage1.GetOcx());
			}

		}

		//画像ファイル保存
		private void btnSaveImage_Click(object sender, EventArgs e)
		{
			saveFileDialog1.InitialDirectory = m_strDirectory;
			saveFileDialog1.Filter = "bitmap file(*.bmp)|*.bmp";
			saveFileDialog1.DefaultExt = "*.bmp";
			saveFileDialog1.ShowDialog();

			if (saveFileDialog1.FileName.Length > 0)
			{
				//FileAccessコンポーネントを使って
				//Imageをファイルに保存します｡
				axFvxFileAccess1.FilePath = saveFileDialog1.FileName;
				try
				{
					axFvxFileAccess1.SaveBitmap(axFvxImage1.GetOcx());
				}
				catch (Exception)
				{
					int IntErr = 0;
					axFvxFileAccess1.GetErrReport(out IntErr);
					MessageBox.Show("保存に失敗しました。　ErrCode= " + IntErr.ToString());
					return;
				}
			}
		}

		//検出円設定
		private void btnSetEdgeScanCircle_Click(object sender, EventArgs e)
		{
			int intOK;
			//設定ダイアログの初期値をセット
			axFvxROI1.SetCircle(m_EdgeScanCircle);

			//FvxROIを使って検出円選択ダイアログを出します｡
			axFvxROI1.UseCircle = true;
			axFvxROI1.UsePoint = false;
			axFvxROI1.UseRectangle = false;
			axFvxROI1.MakeROIDialog(axFvxImage1.GetOcx(),out intOK);

			if (intOK == 1 && axFvxROI1.CheckCircle == 1)
			{
				axFvxROI1.GetCircle(m_EdgeScanCircle);
			}
			ApplyChanges();
			DrawEdgeScanLine();
		}

		//パラメータ設定
		private bool ApplyChanges()
		{
			int Temp;

			//しきい値
			if(int.TryParse(txtThreshold.Text,out Temp) == false || Temp < 0 || Temp > 200)
			{
				MessageBox.Show("閾値は 0～200 の範囲で指定して下さい", "パラメータエラー", MessageBoxButtons.OK, MessageBoxIcon.Error);
				return(false);
			}
			axFvxEdge1.Thresh = Temp;

			//片幅
			if (int.TryParse(txtLineWidth.Text, out Temp) == false || Temp < 0)
			{
				MessageBox.Show("片幅は 0 以上で指定して下さい", "パラメータエラー", MessageBoxButtons.OK, MessageBoxIcon.Error);
				return(false);
			}
			axFvxEdge1.LineWidth = Temp;

			//検出線数
			if (int.TryParse(txtLineNum.Text, out Temp) == false || Temp < 1 || Temp > 360)
			{
				MessageBox.Show("検出線数は 1～360 の範囲で指定して下さい", "パラメータエラー", MessageBoxButtons.OK, MessageBoxIcon.Error);
				return(false);
			}
			axFvxEdge1.LineNum = Temp;

			axFvxEdge1.EdgeColor = (FVXCOMEDGELib.EDGE_COLOR_TYPE)cmbEdgeColor.SelectedIndex;			//エッジ極性
			axFvxEdge1.StartPosition = (FVXCOMEDGELib.START_POSITION_TYPE)cmbEdgeStartPos.SelectedIndex;	//検出方向
			axFvxEdge1.EdgePoint = (FVXCOMEDGELib.EDGE_POINT_TYPE)cmbEdgePoint.SelectedIndex;			//エッジ位置

			return(true);
		}

		//検出円とその中心から伸びる検出線の描画
		private void DrawEdgeScanLine()
		{
			FvLineSegment pLineSeg = new FvLineSegment();
			double dblRad;

			axFvxDisplayEx1.ClearOverlay(-1);

			//エッジ検出線を描画
			axFvxDisplayEx1.PenColor = Color.Pink;
			axFvxDisplayEx1.DrawData(0, m_EdgeScanCircle);
			pLineSeg.xs = m_EdgeScanCircle.x;
			pLineSeg.ys = m_EdgeScanCircle.y;
			dblRad = 2 * PI / axFvxEdge1.LineNum;	//単位はラジアン
			for (int i = 0; i < axFvxEdge1.LineNum; i++ )
			{
				pLineSeg.xe = pLineSeg.xs + Math.Cos(dblRad * i) * m_EdgeScanCircle.radius;
				pLineSeg.ye = pLineSeg.ys + Math.Sin(dblRad * i) * m_EdgeScanCircle.radius;
				axFvxDisplayEx1.DrawData(0, pLineSeg);
			}
		}

		//表示更新
		private void btnReDraw_Click(object sender, EventArgs e)
		{
			if (ApplyChanges() == true)
				DrawEdgeScanLine();
		}

		//実行
		private void btnExecute_Click(object sender, EventArgs e)
		{
			FvCircle pCircle = new FvCircle();
			FvPointGroup pPointGroup = new FvPointGroup();
			FvLineSegment pLineSeg = new FvLineSegment();
			double dblWeight; //ロバスト推定重み
			//重みは、データのバラつきの標準偏差の1～3倍を目安としてください。

			double dblTime = 0;
			int intCount = 0;
			int intErr = 0;

			FvPoint pOrigin = new FvPoint();	//エッジとの距離を測る点（円の中心）
			double dblDistance = 0;				//距離計算用バッファ
			double dblMaxDistance = 0;			//最大距離
			double dblMinDistance = 0;			//最小距離
			int intMaxDistNo = 0;				//最大距離となる点の番号
			int intMinDistNo = 0;				//最小距離となる点の番号

			axFvxDisplayEx1.ClearOverlay(-1);
			txtResult.Text = "";

			//パラメータにエラーがあったら計測せず終了
			if (ApplyChanges() == false) return;
			if (double.TryParse(txtWeight.Text, out dblWeight) == false)
			{
				MessageBox.Show("重みの値が不正です", "パラメータエラー", MessageBoxButtons.OK, MessageBoxIcon.Error);
				return ;
			}
			
			//検出円を描画
			axFvxDisplayEx1.PenColor = Color.Pink;
			axFvxDisplayEx1.DrawData(0,m_EdgeScanCircle);
			DrawCrossMark(m_EdgeScanCircle.x, m_EdgeScanCircle.y);
			
			//時間計測開始
			axFvxPerformanceCounter1.StartCount();

			//エッジ検出実行
			try{
				axFvxEdge1.Execute(axFvxImage1.GetOcx(), m_EdgeScanCircle, out intCount);
			}
			catch(Exception){
				axFvxEdge1.GetErrReport( out intErr);
				MessageBox.Show(GetEdgeErrMesg(intErr),"エッジ取得エラー", MessageBoxButtons.OK,MessageBoxIcon.Error);
				return;
			}

			//エッジと円は青で描画
			axFvxDisplayEx1.PenColor = Color.Blue;

			//検出されたエッジ点座標の取得
			axFvxEdge1.GetResult(0,pPointGroup);

			//ロバスト推定のためにエッジ点座標をFvxCalcDataにセット
			try
			{
				axFvxCalcData1.SetPoint(0, pPointGroup);
				for (int i = 0; i < pPointGroup.Count; i++)
					DrawCrossMark(pPointGroup.Item2(i).x, pPointGroup.Item2(i).y); //エッジ点描画

				//3つ以上のエッジがあったら円のロバスト推定を実行
				if (pPointGroup.Count > 2)
				{
					axFvxCalcData1.CalcCircle_Robust(pPointGroup.Count, dblWeight);
					axFvxCalcData1.CalcCircle_GetResult(pCircle);
					axFvxDisplayEx1.DrawData(0, pCircle);
					DrawCrossMark(pCircle.x, pCircle.y);

					txtResult.Text = string.Format("円中心：( {0:0.0} , {1:0.0} )\r\n" +
													"円半径： {2:0.0}\r\n", pCircle.x, pCircle.y, pCircle.radius);
				}
				else
				{
					MessageBox.Show("エッジが少なすぎます", "エッジ検出");
					return;
				}
			}
			catch (System.Runtime.InteropServices.COMException)//COMのエラーを取得
			{
				axFvxCalcData1.GetErrReport(out intErr);
				MessageBox.Show(GetCalcDataErrMesg(intErr), "ロバスト推定エラー", MessageBoxButtons.OK, MessageBoxIcon.Error);
				return;
			}
			catch(SystemException ex)//システムエラー
			{
				MessageBox.Show(ex.Message, "システムエラー", MessageBoxButtons.OK, MessageBoxIcon.Error);
				return;
			}

			//どの円の中心からの距離を測るか
			if (cmbOrigin.SelectedIndex == 0)
			{	//検出円からの中心
				pOrigin.x = m_EdgeScanCircle.x;
				pOrigin.y = m_EdgeScanCircle.y;
			}
			else //エッジから求めた点の中心
			{
				pOrigin.x = pCircle.x;
				pOrigin.y = pCircle.y;
			}

			//1番目のエッジ点と円の中心との距離を初期値として設定
			dblDistance = Math.Sqrt(Math.Pow((pOrigin.x - pPointGroup.Item2(0).x),2) + Math.Pow((pOrigin.y - pPointGroup.Item2(0).y),2));
			dblMaxDistance = dblDistance;
			dblMinDistance = dblDistance;

			for( int i = 0; i < pPointGroup.Count; i++)
			{	//円の中心とエッジ点との距離を計算
				dblDistance = Math.Sqrt(Math.Pow((pOrigin.x - pPointGroup.Item2(i).x), 2) + Math.Pow((pOrigin.y - pPointGroup.Item2(i).y),2));
				//最小値・最大値のピックアップ
				if(dblMaxDistance < dblDistance)
				{
					dblMaxDistance = dblDistance;
					intMaxDistNo = i;
				}
				if(dblMinDistance > dblDistance)
				{
					dblMinDistance = dblDistance;
					intMinDistNo = i;
				}
			}

			axFvxDisplayEx1.PenColor = Color.Red;

			//中心から最大距離点までを描画
			pLineSeg.xs = pOrigin.x;
			pLineSeg.ys = pOrigin.y;
			pLineSeg.xe = pPointGroup.Item2(intMaxDistNo).x;
			pLineSeg.ye = pPointGroup.Item2(intMaxDistNo).y;
			axFvxDisplayEx1.DrawData(0, pLineSeg);
			
			//中心から最小距離点までを描画
			pLineSeg.xe = pPointGroup.Item2(intMinDistNo).x;
			pLineSeg.ye = pPointGroup.Item2(intMinDistNo).y;
			axFvxDisplayEx1.DrawData(0, pLineSeg);
			
			txtResult.Text = txtResult.Text + string.Format("最小距離：{0:0.00}\r\n最大距離：{1:0.00}",dblMinDistance,dblMaxDistance);

			//時間計測終了
			axFvxPerformanceCounter1.StopCount(out dblTime);
			txtResult.Text = string.Format("全時間処理：{0:0.0}ms\r\n\r\n"+txtResult.Text,dblTime);

		}

		private void DrawCrossMark(double dbl_x, double dbl_y)
		{
			FvLineSegment pLineSeg = new FvLineSegment();

			pLineSeg.xs = dbl_x - 5;
			pLineSeg.ys = dbl_y - 5;
			pLineSeg.xe = dbl_x + 5;
			pLineSeg.ye = dbl_y + 5;
			axFvxDisplayEx1.DrawData(0, pLineSeg);

			pLineSeg.xs = dbl_x + 5;
			pLineSeg.ys = dbl_y - 5;
			pLineSeg.xe = dbl_x - 5;
			pLineSeg.ye = dbl_y + 5;
			axFvxDisplayEx1.DrawData(0, pLineSeg);
		}

		private string GetEdgeErrMesg(int intErr)
		{
			string Str = "";

			switch (intErr)
			{
				case 4005:
					Str = "パラメータが違います";
					break;
				case 4006:
					Str = "引数の型が違います";
					break;
				case 4007:
					Str = "引数にNULLが指定されています";
					break;
				case 4013:
					Str = "処理領域が指定されていません";
					break;
			}
			return (Str);
		}

		private string GetCalcDataErrMesg(int intErr)
		{
			string Str = "";

			switch (intErr)
			{
				case 1605:
					Str = "パラメータが違います";
					break;
				case 1606:
					Str = "引数の型が違います";
					break;
				case 1607:
					Str = "引数にNULLが指定されています";
					break;
				case 1610:
					Str = "メモリの確保に失敗しました";
					break;
				case 1650:
					Str = "計算不可能です";
					break;
			}
			return (Str);
		}

		private void txtThreshold_KeyPress(object sender, KeyPressEventArgs e)
		{
			if ((e.KeyChar < '0' || e.KeyChar > '9') && e.KeyChar != '\b')
				e.Handled = true;
		}
		private void txtLineWidth_KeyPress(object sender, KeyPressEventArgs e)
		{
			if ((e.KeyChar < '0' || e.KeyChar > '9') && e.KeyChar != '\b')
				e.Handled = true;
		}
		private void txtLineNum_KeyPress(object sender, KeyPressEventArgs e)
		{
			if ((e.KeyChar < '0' || e.KeyChar > '9') && e.KeyChar != '\b')
				e.Handled = true;
		}

		private void txtWeight_KeyPress(object sender, KeyPressEventArgs e)
		{
			if ((e.KeyChar < '0' || e.KeyChar > '9') && e.KeyChar != '\b' && e.KeyChar != '.')
				e.Handled = true;
		}


	}
}